@@ -20,10 +20,9 @@ package org.apache.cassandra.tools.nodetool.stats;
 
 import java.util.*;
 
-import javax.management.InstanceNotFoundException;
-
 import com.google.common.collect.ArrayListMultimap;
 
+import javax.management.InstanceNotFoundException;
 import org.apache.cassandra.db.*;
 import org.apache.cassandra.io.util.*;
 import org.apache.cassandra.metrics.*;
@@ -33,16 +32,43 @@ public class TableStatsHolder implements StatsHolder
 {
     public final List<StatsKeyspace> keyspaces;
     public final int numberOfTables;
+    public final boolean humanReadable;
+    public final String sortKey;
+    public final int top;
+    public final boolean locationCheck;
 
-    public TableStatsHolder(NodeProbe probe, boolean humanReadable, boolean ignore, List<String> tableNames)
+    public TableStatsHolder(NodeProbe probe, boolean humanReadable, boolean ignore, List<String> tableNames, String sortKey, int top, boolean locationCheck)
     {
         this.keyspaces = new ArrayList<>();
-        this.numberOfTables = probe.getNumberOfTables();
-        this.initializeKeyspaces(probe, humanReadable, ignore, tableNames);
+        this.humanReadable = humanReadable;
+        this.sortKey = sortKey;
+        this.top = top;
+        this.locationCheck = locationCheck;
+
+        if (!this.isTestTableStatsHolder())
+        {
+            this.numberOfTables = probe.getNumberOfTables();
+            this.initializeKeyspaces(probe, ignore, tableNames);
+        }
+        else
+        {
+            this.numberOfTables = 0;
+        }
     }
 
     @Override
     public Map<String, Object> convert2Map()
+    {
+        if (sortKey.isEmpty())
+            return convertAllToMap();
+        else
+            return convertSortedFilteredSubsetToMap();
+    }
+
+    /**
+     * @returns Map<String, Object> a nested HashMap of keyspaces, their tables, and the tables' statistics.
+     */
+    private Map<String, Object> convertAllToMap()
     {
         HashMap<String, Object> mpRet = new HashMap<>();
         mpRet.put("total_number_of_tables", numberOfTables);
@@ -62,51 +88,8 @@ public class TableStatsHolder implements StatsHolder
             Map<String, Map<String, Object>> mpTables = new HashMap<>();
             for (StatsTable table : tables)
             {
-                Map<String, Object> mpTable = new HashMap<>();
-
-                mpTable.put("sstables_in_each_level", table.sstablesInEachLevel);
-                mpTable.put("space_used_live", table.spaceUsedLive);
-                mpTable.put("space_used_total", table.spaceUsedTotal);
-                mpTable.put("space_used_by_snapshots_total", table.spaceUsedBySnapshotsTotal);
-                if (table.offHeapUsed)
-                    mpTable.put("off_heap_memory_used_total", table.offHeapMemoryUsedTotal);
-                mpTable.put("sstable_compression_ratio", table.sstableCompressionRatio);
-                mpTable.put("number_of_partitions_estimate", table.numberOfPartitionsEstimate);
-                mpTable.put("memtable_cell_count", table.memtableCellCount);
-                mpTable.put("memtable_data_size", table.memtableDataSize);
-                if (table.memtableOffHeapUsed)
-                    mpTable.put("memtable_off_heap_memory_used", table.memtableOffHeapMemoryUsed);
-                mpTable.put("memtable_switch_count", table.memtableSwitchCount);
-                mpTable.put("local_read_count", table.localReadCount);
-                mpTable.put("local_read_latency_ms", String.format("%01.3f", table.localReadLatencyMs));
-                mpTable.put("local_write_count", table.localWriteCount);
-                mpTable.put("local_write_latency_ms", String.format("%01.3f", table.localWriteLatencyMs));
-                mpTable.put("pending_flushes", table.pendingFlushes);
-                mpTable.put("percent_repaired", table.percentRepaired);
-                mpTable.put("bloom_filter_false_positives", table.bloomFilterFalsePositives);
-                mpTable.put("bloom_filter_false_ratio", String.format("%01.5f", table.bloomFilterFalseRatio));
-                mpTable.put("bloom_filter_space_used", table.bloomFilterSpaceUsed);
-                if (table.bloomFilterOffHeapUsed)
-                    mpTable.put("bloom_filter_off_heap_memory_used", table.bloomFilterOffHeapMemoryUsed);
-                if (table.indexSummaryOffHeapUsed)
-                    mpTable.put("index_summary_off_heap_memory_used", table.indexSummaryOffHeapMemoryUsed);
-                if (table.compressionMetadataOffHeapUsed)
-                    mpTable.put("compression_metadata_off_heap_memory_used",
-                                table.compressionMetadataOffHeapMemoryUsed);
-                mpTable.put("compacted_partition_minimum_bytes", table.compactedPartitionMinimumBytes);
-                mpTable.put("compacted_partition_maximum_bytes", table.compactedPartitionMaximumBytes);
-                mpTable.put("compacted_partition_mean_bytes", table.compactedPartitionMeanBytes);
-                mpTable.put("average_live_cells_per_slice_last_five_minutes",
-                            table.averageLiveCellsPerSliceLastFiveMinutes);
-                mpTable.put("maximum_live_cells_per_slice_last_five_minutes",
-                            table.maximumLiveCellsPerSliceLastFiveMinutes);
-                mpTable.put("average_tombstones_per_slice_last_five_minutes",
-                            table.averageTombstonesPerSliceLastFiveMinutes);
-                mpTable.put("maximum_tombstones_per_slice_last_five_minutes",
-                            table.maximumTombstonesPerSliceLastFiveMinutes);
-                mpTable.put("dropped_mutations", table.droppedMutations);
-
-                mpTables.put(table.name, mpTable);
+                Map<String, Object> mpTable = convertStatsTableToMap(table);
+                mpTables.put(table.tableName, mpTable);
             }
             mpKeyspace.put("tables", mpTables);
             mpRet.put(keyspace.name, mpKeyspace);
@@ -114,7 +97,76 @@ public class TableStatsHolder implements StatsHolder
         return mpRet;
     }
 
-    private void initializeKeyspaces(NodeProbe probe, boolean humanReadable, boolean ignore, List<String> tableNames)
+    /**
+     * @returns Map<String, Object> a nested HashMap of the sorted and filtered table names and the HashMaps of their statistics.
+     */
+    private Map<String, Object> convertSortedFilteredSubsetToMap()
+    {
+        HashMap<String, Object> mpRet = new HashMap<>();
+        mpRet.put("total_number_of_tables", numberOfTables);
+        List<StatsTable> sortedFilteredTables = getSortedFilteredTables();
+        for (StatsTable table : sortedFilteredTables)
+        {
+            String tableDisplayName = table.keyspaceName + "." + table.tableName;
+            Map<String, Object> mpTable = convertStatsTableToMap(table);
+            mpRet.put(tableDisplayName, mpTable);
+        }
+        return mpRet;
+    }
+
+    private Map<String, Object> convertStatsTableToMap(StatsTable table)
+    {
+        Map<String, Object> mpTable = new HashMap<>();
+        mpTable.put("sstables_in_each_level", table.sstablesInEachLevel);
+        mpTable.put("space_used_live", table.spaceUsedLive);
+        mpTable.put("space_used_total", table.spaceUsedTotal);
+        mpTable.put("space_used_by_snapshots_total", table.spaceUsedBySnapshotsTotal);
+        if (table.offHeapUsed)
+            mpTable.put("off_heap_memory_used_total", table.offHeapMemoryUsedTotal);
+        mpTable.put("sstable_compression_ratio", table.sstableCompressionRatio);
+        mpTable.put("number_of_partitions_estimate", table.numberOfPartitionsEstimate);
+        mpTable.put("memtable_cell_count", table.memtableCellCount);
+        mpTable.put("memtable_data_size", table.memtableDataSize);
+        if (table.memtableOffHeapUsed)
+            mpTable.put("memtable_off_heap_memory_used", table.memtableOffHeapMemoryUsed);
+        mpTable.put("memtable_switch_count", table.memtableSwitchCount);
+        mpTable.put("local_read_count", table.localReadCount);
+        mpTable.put("local_read_latency_ms", String.format("%01.3f", table.localReadLatencyMs));
+        mpTable.put("local_write_count", table.localWriteCount);
+        mpTable.put("local_write_latency_ms", String.format("%01.3f", table.localWriteLatencyMs));
+        mpTable.put("pending_flushes", table.pendingFlushes);
+        mpTable.put("percent_repaired", table.percentRepaired);
+        mpTable.put("bytes_repaired", table.bytesRepaired);
+        mpTable.put("bytes_unrepaired", table.bytesUnrepaired);
+        mpTable.put("bytes_pending_repair", table.bytesPendingRepair);
+        mpTable.put("bloom_filter_false_positives", table.bloomFilterFalsePositives);
+        mpTable.put("bloom_filter_false_ratio", String.format("%01.5f", table.bloomFilterFalseRatio));
+        mpTable.put("bloom_filter_space_used", table.bloomFilterSpaceUsed);
+        if (table.bloomFilterOffHeapUsed)
+            mpTable.put("bloom_filter_off_heap_memory_used", table.bloomFilterOffHeapMemoryUsed);
+        if (table.indexSummaryOffHeapUsed)
+            mpTable.put("index_summary_off_heap_memory_used", table.indexSummaryOffHeapMemoryUsed);
+        if (table.compressionMetadataOffHeapUsed)
+            mpTable.put("compression_metadata_off_heap_memory_used",
+                        table.compressionMetadataOffHeapMemoryUsed);
+        mpTable.put("compacted_partition_minimum_bytes", table.compactedPartitionMinimumBytes);
+        mpTable.put("compacted_partition_maximum_bytes", table.compactedPartitionMaximumBytes);
+        mpTable.put("compacted_partition_mean_bytes", table.compactedPartitionMeanBytes);
+        mpTable.put("average_live_cells_per_slice_last_five_minutes",
+                    table.averageLiveCellsPerSliceLastFiveMinutes);
+        mpTable.put("maximum_live_cells_per_slice_last_five_minutes",
+                    table.maximumLiveCellsPerSliceLastFiveMinutes);
+        mpTable.put("average_tombstones_per_slice_last_five_minutes",
+                    table.averageTombstonesPerSliceLastFiveMinutes);
+        mpTable.put("maximum_tombstones_per_slice_last_five_minutes",
+                    table.maximumTombstonesPerSliceLastFiveMinutes);
+        mpTable.put("dropped_mutations", table.droppedMutations);
+        if (locationCheck)
+            mpTable.put("sstables_in_correct_location", table.isInCorrectLocation);
+        return mpTable;
+    }
+
+    private void initializeKeyspaces(NodeProbe probe, boolean ignore, List<String> tableNames)
     {
         OptionFilter filter = new OptionFilter(ignore, tableNames);
         ArrayListMultimap<String, ColumnFamilyStoreMBean> selectedTableMbeans = ArrayListMultimap.create();
@@ -160,9 +212,13 @@ public class TableStatsHolder implements StatsHolder
             {
                 String tableName = table.getTableName();
                 StatsTable statsTable = new StatsTable();
-                statsTable.name = tableName;
+                statsTable.fullName = keyspaceName + "." + tableName;
+                statsTable.keyspaceName = keyspaceName;
+                statsTable.tableName = tableName;
                 statsTable.isIndex = tableName.contains(".");
                 statsTable.sstableCount = probe.getColumnFamilyMetric(keyspaceName, tableName, "LiveSSTableCount");
+                statsTable.oldSSTableCount = probe.getColumnFamilyMetric(keyspaceName, tableName, "OldVersionSSTableCount");
+
                 int[] leveledSStables = table.getSSTableCountPerLevel();
                 if (leveledSStables != null)
                 {
@@ -179,12 +235,18 @@ public class TableStatsHolder implements StatsHolder
                     }
                 }
 
+                if (locationCheck)
+                    statsTable.isInCorrectLocation = !table.hasMisplacedSSTables();
+
                 Long memtableOffHeapSize = null;
                 Long bloomFilterOffHeapSize = null;
                 Long indexSummaryOffHeapSize = null;
                 Long compressionMetadataOffHeapSize = null;
                 Long offHeapSize = null;
                 Double percentRepaired = null;
+                Long bytesRepaired = null;
+                Long bytesUnrepaired = null;
+                Long bytesPendingRepair = null;
 
                 try
                 {
@@ -194,6 +256,9 @@ public class TableStatsHolder implements StatsHolder
                     compressionMetadataOffHeapSize = (Long) probe.getColumnFamilyMetric(keyspaceName, tableName, "CompressionMetadataOffHeapMemoryUsed");
                     offHeapSize = memtableOffHeapSize + bloomFilterOffHeapSize + indexSummaryOffHeapSize + compressionMetadataOffHeapSize;
                     percentRepaired = (Double) probe.getColumnFamilyMetric(keyspaceName, tableName, "PercentRepaired");
+                    bytesRepaired = (Long) probe.getColumnFamilyMetric(keyspaceName, tableName, "BytesRepaired");
+                    bytesUnrepaired = (Long) probe.getColumnFamilyMetric(keyspaceName, tableName, "BytesUnrepaired");
+                    bytesPendingRepair = (Long) probe.getColumnFamilyMetric(keyspaceName, tableName, "BytesPendingRepair");
                 }
                 catch (RuntimeException e)
                 {
@@ -215,6 +280,11 @@ public class TableStatsHolder implements StatsHolder
                 {
                     statsTable.percentRepaired = Math.round(100 * percentRepaired) / 100.0;
                 }
+
+                statsTable.bytesRepaired = bytesRepaired != null ? bytesRepaired : 0;
+                statsTable.bytesUnrepaired = bytesUnrepaired != null ? bytesUnrepaired : 0;
+                statsTable.bytesPendingRepair = bytesPendingRepair != null ? bytesPendingRepair : 0;
+
                 statsTable.sstableCompressionRatio = probe.getColumnFamilyMetric(keyspaceName, tableName, "CompressionRatio");
                 Object estimatedPartitionCount = probe.getColumnFamilyMetric(keyspaceName, tableName, "EstimatedPartitionCount");
                 if (Long.valueOf(-1L).equals(estimatedPartitionCount))
@@ -286,6 +356,24 @@ public class TableStatsHolder implements StatsHolder
         return humanReadable ? FileUtils.stringifyFileSize(bytes) : Long.toString(bytes);
     }
 
+    /**
+     * Sort and filter this TableStatHolder's tables as specified by its sortKey and top attributes.
+     */
+    public List<StatsTable> getSortedFilteredTables() {
+        List<StatsTable> tables = new ArrayList<>();
+        for (StatsKeyspace keyspace : keyspaces)
+            tables.addAll(keyspace.tables);
+        Collections.sort(tables, new StatsTableComparator(sortKey, humanReadable));
+        int k = (tables.size() >= top) ? top : tables.size();
+        if (k > 0)
+            tables = tables.subList(0, k);
+        return tables;
+    }
+
+    protected boolean isTestTableStatsHolder() {
+        return false;
+    }
+
     /**
      * Used for filtering keyspaces and tables to be displayed using the tablestats command.
      */
